/*
 * Bank.java
 * 
 * $Id: Bank.java,v 1.9 2012-05-16 06:18:10 jbf9160 Exp $
 * 
 * $Log: Bank.java,v $
 * Revision 1.9  2012-05-16 06:18:10  jbf9160
 * Comments added.
 *
 * Revision 1.8  2012-05-15 21:09:50  jbf9160
 * BatchMode method created.
 * BatchMode working.
 * Minor updates.
 *
 * Revision 1.7  2012-05-15 11:12:10  jbf9160
 * Small updates.
 * MVC Fully working.
 *
 * Revision 1.6  2012-05-15 10:52:05  jbf9160
 * Interface fully working.
 *
 * Revision 1.5  2012-05-15 07:16:26  jbf9160
 * Bank Model and GUI now can test for an account ID.
 *
 * Revision 1.4  2012-05-15 06:48:26  jbf9160
 * Added headers... hopefully not to late.
 *
 */

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Observable;

/**
 * Bank is the model for the bank GUI interface.
 * also contains the main method for the Bank program.
 * 
 * @author James Fator (jbf9160@rit.edu)
 */
public class Bank extends Observable {
	
	private ArrayList<Account> accounts;
	private String bankFile;

	/**
	 * Constructor for the bank object
	 * 
	 * @param accounts the open accounts from bankFile
	 * @param bankFile the location to the bankFile
	 */
	public Bank(ArrayList<Account> accounts, String bankFile) {
		this.accounts = accounts;
		this.bankFile = bankFile;
	}
	
	/**
	 * main method for bank. Parses the bankFile for accounts and
	 * either opens the BankGUI or runs the batch mode.
	 * 
	 * @param args bankFile [batchFile]
	 */
	public static void main(String[] args) {
		try {
			if (args.length == 1) {
				ArrayList<Account> accounts = getAccounts(args[0]);
				Bank bank = new Bank(accounts, args[0]);
				BankGUI bi = new BankGUI(bank);
				bank.addObserver(bi);
			} else if (args.length == 2) {
				ArrayList<Account> accounts = getAccounts(args[0]);
				Bank bank = new Bank(accounts, args[0]);
				bank.batchMode(args[1]);
			} else {
				System.err.println("Usage: java Bank bankFile [batchFile]");
				return;
			}
		} catch (FileNotFoundException FNFE) {
			System.err.println(args[0] + " does not exist");
			return;
		} catch (IOException IOE) {
			System.err.println("I/O Error - please check bank file");
			return;
		}
	}
	
	/**
	 * getAccounts is a getter for the Account ArrayList
	 * 
	 * @return accounts the open accounts from the bankFile
	 */
	/*Protected*/ ArrayList<Account> getAccounts() {
		return this.accounts;
	}
	
	/**
	 * validId checks if the passed integer is an id to an open Account
	 * 
	 * @param id id number of the Account
	 * @return true if there is an Account with that id
	 */
	/*Protected*/ boolean validId(int id) {
		for (Account u : this.accounts) {
			if (u.getId() == id) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * validPin checks if the given pin works with the given id
	 * 
	 * @param id the Account to check
	 * @param pin the pin number that is being tested
	 * @return true if the Account can be opened with said pin
	 * 
	 * @assumed id exists
	 */
	/*Protected*/ boolean validPin(int id, int pin) {
		for (Account u : this.accounts) {
			if (u.getId() == id) {
				if (u.checkPin(pin)) {
					return true;
				} else {
					return false;
				}
			}
		}
		return false;
	}
	
	/**
	 * getBal returns the balance for the requested Account
	 * 
	 * @param id the Account to look up the balance for
	 * @param pin the pin for said Account
	 * @return the balance of the given Account
	 * 
	 * @assumed id exists and pin is correct
	 */
	/*Protected*/ double getBal(int id, int pin) {
		for (Account u : this.accounts) {
			if (u.getId() == id) {
				if (u.checkPin(pin)) {
					return u.getBalance();
				} else {
					return 0.0;
				}
			}
		}
		return 0.0;
	}
	
	/**
	 * deposit updates the Account with an increase in value
	 * 
	 * @param id id of the Account
	 * @param pin pin for the Account
	 * @param bal amount to be deposited
	 * 
	 * @assumed id exists and pin is correct
	 */
	/*Package*/ void deposit(int id, int pin, double bal) {
		for (Account u : this.accounts) {
			if (u.getId() == id) {
				if (u.checkPin(pin) && bal != -1) {
					u.deposit(bal);
					this.setChanged();
					this.notifyObservers();
				}
			}
		}
	}
	
	/**
	 * withdraw updates the Account with a decrease in value
	 * 
	 * @param id id of the Account
	 * @param pin pin for the Account
	 * @param bal amount to be withdrawn
	 * @return true if the money could be withdrawn
	 * 
	 * @assumed id exists and pin is correct
	 */
	/*Package*/ boolean withdraw(int id, int pin, double bal) {
		for (Account u : this.accounts) {
			if (u.getId() == id) {
				if (u.checkPin(pin) && bal != -1 && u.getBalance() > bal) {
					u.withdraw(bal);
					this.setChanged();
					this.notifyObservers();
					return true;
				} else {
					return false;
				}
			}
		}
		return false;
	}
	
	/**
	 * updateAccount changes the given account into a different type
	 * 
	 * @param account Account to change
	 * @param type 'c' for checking, 's' for saving, or 'd' for cd
	 * @return index of the Account in the ArrayList
	 */
	/*Protected*/ int updateAccount(Account account, char type) {
		int index = accounts.indexOf(account);
		Account newAccount = account.phoenix(type);
		this.accounts.remove(account);
		this.accounts.add(index, newAccount);
		return index;
	}
	
	private static ArrayList<Account> getAccounts(String bankFile) 
					throws IOException {

		BufferedReader in = new BufferedReader(new FileReader(bankFile));
		ArrayList<Account> accounts = new ArrayList<Account>();
		String[] args;
		String line;
		while ((line = in.readLine()) != null) {
			args = line.split("-");
			if (args[0].equals("d")) {
				accounts.add(new CertificateOfDeposit('d',
									  Integer.parseInt(args[1]),
									  Integer.parseInt(args[2]),
									  Double.parseDouble(args[3])));
			} else if (args[0].equals("c")) {
				accounts.add(new Checking('c',
						  Integer.parseInt(args[1]),
						  Integer.parseInt(args[2]),
						  Double.parseDouble(args[3])));
			} else if (args[0].equals("s")) {
				accounts.add(new Saving('s',
						  Integer.parseInt(args[1]),
						  Integer.parseInt(args[2]),
						  Double.parseDouble(args[3])));
			}
		}
		return accounts;
	}
	
	private void batchMode(String batchFile) {
		try {
			System.out.println("============== Initial Bank Data " +
					"==============\n");
			System.out.println("Account Type\tAccount\tBalance");
			System.out.println("____________\t_______\t_______\n");
			for (Account a : accounts) {
				System.out.println(a.prettyPrint() + "\t"
						+ a.getId() + "\t$\t" + a.getBalance());
			}
			System.out.println("=================================" +
					"==============\n");
			BufferedReader in = new BufferedReader(new FileReader(batchFile));
			String c;
			String[] line;
			while ((c = in.readLine()) != null) {
				line = c.split(" ");
				if (line[0].equals("o")) {
					if (line[1].equals("x")) {
						if (Integer.parseInt(line[4]) >= 500) {
							System.out.println(
									line[2]+"\to\tx\tOpen: Success\t$\t"+
												line[4]);
							this.accounts.add(new Checking('c',
									  Integer.parseInt(line[2]),
									  Integer.parseInt(line[3]),
									  Double.parseDouble(line[4])));
						} else {
							System.out.println(
									line[2]+"\to\tx\tOpen: Failed");
						}
					} else if (line[1].equals("c")) {
						if (Integer.parseInt(line[4]) >= 50) {
							System.out.println(
									line[2]+"\to\tc\tOpen: Success\t$\t"+
												line[4]);
							this.accounts.add(new CertificateOfDeposit('d',
									  Integer.parseInt(line[2]),
									  Integer.parseInt(line[3]),
									  Double.parseDouble(line[4])));
						} else {
							System.out.println(
									line[2]+"\to\tc\tOpen: Failed");
						}
					} else if (line[1].equals("s")) {
						if (Integer.parseInt(line[4]) >= 50) {
							System.out.println(
									line[2]+"\to\ts\tOpen: Success\t$\t"+
												line[4]);
							this.accounts.add(new Saving('s',
									  Integer.parseInt(line[2]),
									  Integer.parseInt(line[3]),
									  Double.parseDouble(line[4])));
						} else {
							System.out.println(
									line[2]+"\to\ts\tOpen: Failed");
						}
					}
				} else if (line[0].equals("c")) {
					this.closeAccount(Integer.parseInt(line[1]));
				} else if (line[0].equals("d")) {
					int temp = 0;
					Double bal = Double.parseDouble(line[2]);
					for (Account a : this.accounts) {
						if (a.getId() == Integer.parseInt(line[1])) {
							a.deposit(bal);
							System.out.println(
									line[1]+"\td\t\t$\t"+bal+"\t$\t"+
														a.getBalance());
							temp++;
						}
					}
					if (temp == 0) {
						System.out.println(
								line[1]+"\td\t\t$\t"+bal+"\tFailed");
					}
				} else if (line[0].equals("w")) {
					int temp = 0;
					Double bal = Double.parseDouble(line[2]);
					for (Account a : this.accounts) {
						if (a.getId() == Integer.parseInt(line[1]) &&
								a.getBalance() > bal) {
							a.withdraw(bal);
							System.out.println(
									line[1]+"\tw\t\t$\t"+bal+"\t$\t"+
														a.getBalance());
							temp++;
						}
					}
					if (temp == 0) {
						System.out.println(
								line[1]+"\tw\t\t$\t"+bal+"\tFailed");
					}
				} else {
					System.out.println("\n============== Interest Report " +
							"==============");
					System.out.println("Account Adjustment\tNew Balance");
					System.out.println("_______ __________\t___________");
					for (Account a : this.accounts) {
						if (a.getType()=='d') {
							String intr = Double.toString(
									(double)Math.round(
											a.collectInterest() * 100) / 100);
							System.out.println(a.getId()+
									"\t$     "+intr+"\t$\t"+
									a.getBalance());
						} else if (a.getType()=='c') {
							String pent = Double.toString(
								(double)Math.round(
										a.getPenalty() * 100) / 100);
							System.out.println(a.getId()+
									"\t$     "+pent+"\t$\t"+
									a.getBalance());
						} else if (a.getType()=='s') {
							double intr = a.collectInterest();
							double pent = a.getPenalty();
							String total = Double.toString(
									(double)Math.round(
									(intr-pent) * 100) / 100);
							System.out.println(a.getId()+
									"\t$     "+total+"\t$\t"+
									a.getBalance());
						}
					}
					System.out.println("===============================" +
							"==============");
				}
			}
			this.finalData();
		} catch (FileNotFoundException f) {
			System.err.println("Batch File not found!");
			return;
		} catch (IOException ex) {
			System.err.println("I/O Error - please check batch file");
			return;
		}
	}
	
	/**
	 * Prints the final bank data to System.out
	 */
	/*Package*/ void finalData() {
		try {
			System.out.println("================ Final Bank Data "
					+ "==============\n");
			System.out.println("Account Type\tAccount Balance");
			System.out.println("____________\t_______ _______\n");
			FileWriter out = new FileWriter(this.bankFile);
			out.write(Integer.toString(this.accounts.size())+"\n");
			for (Account a : accounts) {
				System.out.println(a.prettyPrint() + "\t"
						+ a.getId() + "\t$\t" + a.getBalance());
				out.write(a.getType()+"-"+a.getId()+"-"+
						a.getPin()+"-"+a.getBalance()+"\n");
			}
			System.out.println("================================="
					+ "==============\n");
			out.close();
		} catch (IOException e) {
			System.err.println("Unable to write accounts to bank file");
		}
	}
	
	private void closeAccount(int id) {
		for (Account a : accounts) {
			if (a.getId() == id) {
				System.out.println(
						id+"\tc\t\tClosed: Success\t$\t"+
											a.getBalance());
				accounts.remove(a);
				return;
			}
		}
		System.out.println(id+"\tc\t\tClosed: Failed");
	}
}